using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace RobotCtrl
{
    public class MotorCtrl
    {
        protected double nominalSpeed;		// m/s
        double distance;					// m
        double currentSpeed;				// m/s
        double acceleration;				// m/s^2
        double ticks;
        bool run;
        Thread thread;
        int status;

        public virtual int Status { get { return status; } }
        public virtual bool Ready { get { return true; } }
        public virtual double Distance { get { return getDistance(); } set { distance = value; } }	// m
        public virtual double Speed { set { setSpeed(value); } get { return nominalSpeed; } }		// m/s
        public virtual double Acceleration { set { setAcceleration(value); } }						// m/(s*s)
        public virtual double CurrentSpeed { get { return currentSpeed; } }
        public bool Stopped { get { return (Status & 0x80) == 0; } }

        protected virtual int Ticks { get { return (int)Math.Round(ticks); } }

        public MotorCtrl()
        {
            reset();
            ticks = 0;
            run = true;
            if (!(this is MotorCtrl_HW))
            {
                thread = new Thread(simulation);
                thread.Priority = ThreadPriority.AboveNormal;
                thread.Start();
            }
        }

        public virtual void Reset()
        {
            reset();
        }

        public virtual void Go()
        {
            status = 0x80;
        }

        public virtual void Stop()
        {
            status = 0x00;
        }

        public virtual void Close()
        {
            run = false;
            if (thread != null)
                thread.Join();
        }

        public virtual void SetPID(int proportional, int integral, int derivative, int integralLimit, int derivativeInterval)
        {
            // Muss nicht implementiert werden, da Simulation nicht genug genau ist.
        }

        void reset()
        {
            nominalSpeed = 0;
            currentSpeed = 0;
            acceleration = 0.25;
            SetPID(100, 20, 1000, 1000, 1);
        }

        void setSpeed(double speed)
        {
            nominalSpeed = speed;
        }

        void setAcceleration(double accel)
        {
            acceleration = accel;
        }

        double getDistance()
        {
            return Ticks * Config.MeterPerTick;
        }

        void simulation()
        {
            int time = Environment.TickCount;
            while (run)
            {
                if (!Stopped)
                {
                    int dtime = Environment.TickCount - time;
                    time += dtime;
                    double dt = dtime / 1000.0; // Delta t in [s]

                    // Encoder Ticks gemss currentSpeed aktualisieren
                    // [m] = [m/s] * [s] / [m/Tick]
                    ticks += currentSpeed * dt / Config.MeterPerTick;

                    // Mit der Vorgegebenen Beschleunigung currentSpeed an nominalSpeed angleichen
                    // [m/s] = [m/s^2] * [s]
                    var speedDifference = acceleration * dt;
                    if (Math.Abs(nominalSpeed - currentSpeed) <= speedDifference)
                    {
                        currentSpeed = nominalSpeed;
                        // Keep speed
                    }
                    else if (currentSpeed < nominalSpeed)
                    {
                        currentSpeed += speedDifference;
                    }
                    else
                    {
                        currentSpeed -= speedDifference;
                    }

                }
                Thread.Sleep(Config.DefaultLoopWaitTime);
            }
        }
    }
}
