﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Reflection;
using System.Data;
using System.Text;
using System.Windows.Forms;
using RobotCtrl;

namespace RobotView
{
    public partial class WorldView : UserControl
    {

        private RectangleF viewPort;
        private Bitmap bitmapBuffer;    //used for double buffering
        private Timer updateTimer;

        private WorldView()
        {
            InitializeComponent();
        }

        public WorldView(RectangleF viewPort)
            : this()
        {
            this.viewPort = viewPort;
            updateTimer = new Timer();
            updateTimer.Interval = 1000;
            updateTimer.Tick += new EventHandler(updateTimer_Tick);
            updateTimer.Enabled = true;
        }

        void updateTimer_Tick(object sender, EventArgs e)
        {
            this.Invalidate();
        }

        /// <summary>
        /// Drawing objects to the GUI. This method implements double buffering.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnPaint(PaintEventArgs e)
        {
            if (bitmapBuffer == null)
            {
                bitmapBuffer = new Bitmap(this.ClientSize.Width, this.ClientSize.Height);
            }

            Graphics g = Graphics.FromImage(bitmapBuffer);

            //Paint graphics on g here 
            DrawBackground(g);
            DrawObstacleMap(g);
            DrawGrid(g);
            DrawAxes(g);
            DrawRobots(g);

            g.Dispose();

            //Copy the back buffer to the screen 
            e.Graphics.DrawImage(bitmapBuffer, 0, 0);

            //base.OnPaint (e); //optional but not recommended
        }

        /// <summary>
        /// Prohibit the Usercontrol to repaint it's background.
        /// Paint the background manually.
        /// </summary>
        /// <param name="pevent"></param>
        protected override void OnPaintBackground(PaintEventArgs pevent)
        {}

        /// <summary>
        /// Allow the Usercontrol to rescale its graphics.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnResize(EventArgs e)
        {
            if (bitmapBuffer != null)
            {
                bitmapBuffer.Dispose();
                bitmapBuffer = null;
            }
            base.OnResize(e);
        } 

        /// <summary>
        /// Calculates 1) the ratio between the height of this window and the height of viewPort
        /// and 2) the ratio between the width of this window and the width of viewPort respectively.
        /// Finally, it returns the lower of both values. This value can be used to scale all graphical objects.
        /// </summary>
        /// <returns></returns>
        private double GetScaleFactor()
        {
            //get scale factors for height and width of the current form
            double heightScaleFactor = this.Height / (viewPort.Bottom - viewPort.Top);
            double widthScaleFactor = this.Width / (viewPort.Right - viewPort.Left);

            //return the lower scale factor
            if (heightScaleFactor <= widthScaleFactor)
            {
                return heightScaleFactor;
            }
            else
            {
                return widthScaleFactor;
            }
        }

        private int GetVerticalOffset()
        {
            double scaleFactor = GetScaleFactor();
            if (viewPort.Top * scaleFactor < 0)
            {
                return Convert.ToInt32(Math.Abs(viewPort.Top * scaleFactor));
            }
            else
            {
                return 1;
            }
        }

        private int GetHorizontalOffset()
        {
            double scaleFactor = GetScaleFactor();
            if (viewPort.Left * scaleFactor < 0)
            {
                return Convert.ToInt32(Math.Abs(viewPort.Left * scaleFactor));
            }
            else
            {
                return 1;
            }
        }

        /// <summary>
        /// Draw background.
        /// </summary>
        /// <param name="g"></param>
        private void DrawBackground(Graphics g)
        {
            g.Clear(Color.White);
        }

        private void DrawGrid(Graphics g)
        {
            Pen gridPen = new Pen(Color.Gray);
            gridPen.DashStyle = System.Drawing.Drawing2D.DashStyle.Dash;

            double scaleFactor = GetScaleFactor();

            //calculate a certain offset to avoid graphics being painted outside visible area
            int verticalOffset = GetVerticalOffset();
            int horizontalOffset = GetHorizontalOffset();

            //draw vertical grid lines
            for (double x = viewPort.Left; x <= viewPort.Right; x++)
            {
                g.DrawLine(gridPen,
                    Convert.ToInt32(x * scaleFactor + horizontalOffset),
                    Convert.ToInt32(viewPort.Top * scaleFactor + verticalOffset),
                    Convert.ToInt32(x * scaleFactor + horizontalOffset),
                    Convert.ToInt32(viewPort.Bottom * scaleFactor + verticalOffset));
            }

            //draw horizontal grid lines
            for (double y = viewPort.Top; y <= viewPort.Bottom; y++)
            {
                g.DrawLine(gridPen,
                    Convert.ToInt32(viewPort.Left * scaleFactor + horizontalOffset),
                    Convert.ToInt32((y - viewPort.Top) * scaleFactor),
                    Convert.ToInt32(viewPort.Right * scaleFactor + horizontalOffset),
                    Convert.ToInt32((y - viewPort.Top) * scaleFactor));
            }
        }

        /// <summary>
        /// Draw X and Y axes
        /// </summary>
        /// <param name="g"></param>
        private void DrawAxes(Graphics g)
        {
            double scaleFactor = GetScaleFactor();
            int verticalOffset = GetVerticalOffset();
            int horizontalOffset = GetHorizontalOffset();

            //draw Y axis
            g.DrawLine(new Pen(Color.Gray, 2.0f),
                    Convert.ToInt32(Math.Abs(viewPort.Left) * scaleFactor),
                    Convert.ToInt32(0),
                    Convert.ToInt32(Math.Abs(viewPort.Left) * scaleFactor),
                    Convert.ToInt32((Math.Abs(viewPort.Top) + Math.Abs(viewPort.Bottom)) * scaleFactor));

            //draw X axis
            g.DrawLine(new Pen(Color.Gray, 2.0f),
                    Convert.ToInt32(0),
                    Convert.ToInt32(Math.Abs(viewPort.Top) * scaleFactor),
                    Convert.ToInt32((Math.Abs(viewPort.Left) + Math.Abs(viewPort.Right)) * scaleFactor),
                    Convert.ToInt32(Math.Abs(viewPort.Bottom) * scaleFactor));
        }

        /// <summary>
        /// Draw robot(s).
        /// </summary>
        /// <param name="g"></param>
        private void DrawRobots(Graphics g)
        {
            double scaleFactor = GetScaleFactor();
            int verticalOffset = GetVerticalOffset();
            int horizontalOffset = GetHorizontalOffset();

            foreach (Robot robot in World.Robots)
            {
                Pen robotPen = new Pen(robot.BorderColor);
                robotPen.DashStyle = System.Drawing.Drawing2D.DashStyle.Solid;
                robotPen.Width = 2.0f;

                Rectangle rect = new Rectangle(
                    Convert.ToInt32((robot.Position.X * scaleFactor) + horizontalOffset - ((robot.Dimensions.Width * scaleFactor) / 2)),
                    Convert.ToInt32((robot.Position.Y * scaleFactor) + verticalOffset - ((robot.Dimensions.Height * scaleFactor) / 2)),
                    Convert.ToInt32(robot.Dimensions.Height * scaleFactor),
                    Convert.ToInt32(robot.Dimensions.Width * scaleFactor));
                g.DrawEllipse(robotPen, rect);
                g.FillEllipse(new SolidBrush(robot.Color), rect);

                //DEBUG Information about current robot
                //g.DrawString("x=" + robot.Position.X + " y=" + robot.Position.Y, new Font("Arial", 10.0f, FontStyle.Regular), new SolidBrush(Color.White), rect);
                
                //Draw angle line
                Pen anglePen = new Pen(Color.Red);
                anglePen.DashStyle = System.Drawing.Drawing2D.DashStyle.Solid;
                anglePen.Width = 2.0f;

                int anglelineStartX = Convert.ToInt32((robot.Position.X * scaleFactor) + horizontalOffset);
                int anglelineStartY = Convert.ToInt32((robot.Position.Y * scaleFactor) + verticalOffset);
                g.DrawLine(anglePen,
                    anglelineStartX,
                    anglelineStartY,
                    anglelineStartX + Convert.ToInt32(Math.Cos(robot.Position.Angle / 180 * Math.PI) * (robot.Dimensions.Width * scaleFactor / 2)),
                    anglelineStartY + Convert.ToInt32(Math.Sin(robot.Position.Angle / 180 * Math.PI) * (robot.Dimensions.Height * scaleFactor / 2)));

                //Draw radar line
                Pen radarPen = new Pen(Color.LawnGreen);
                radarPen.DashStyle = System.Drawing.Drawing2D.DashStyle.Dash;
                radarPen.Width = 2.0f;

                int antennaStartX = Convert.ToInt32((robot.Position.X * scaleFactor) + horizontalOffset +
                    Math.Cos(robot.Position.Angle * Math.PI / 180) * (robot.Radar.AntennaPosition.X * scaleFactor) +
                    Math.Sin(-robot.Position.Angle * Math.PI / 180) * (robot.Radar.AntennaPosition.Y * scaleFactor));
                int antennaStartY = Convert.ToInt32((robot.Position.Y * scaleFactor) + verticalOffset +
                    Math.Sin(robot.Position.Angle * Math.PI / 180) * (robot.Radar.AntennaPosition.Y * scaleFactor) +
                    Math.Cos(-robot.Position.Angle * Math.PI / 180) * (robot.Radar.AntennaPosition.X * scaleFactor));
                g.DrawLine(radarPen,
                    antennaStartX,
                    antennaStartY,
                    antennaStartX + Convert.ToInt32(Math.Cos((robot.Position.Angle + robot.Radar.AntennaPosition.Angle) / 180 * Math.PI) * (robot.Radar.Distance * scaleFactor)),
                    antennaStartY + Convert.ToInt32(Math.Sin((robot.Position.Angle + robot.Radar.AntennaPosition.Angle) / 180 * Math.PI) * (robot.Radar.Distance * scaleFactor)));

            }
        }

        /// <summary>
        /// Draw obstacle(s).
        /// </summary>
        /// <param name="g"></param>
        private void DrawObstacleMap(Graphics g)
        {
            double scaleFactor = GetScaleFactor();
            int verticalOffset = GetVerticalOffset();
            int horizontalOffset = GetHorizontalOffset();

            foreach (ObstacleMap map in World.Obstacles)
            {
                g.DrawImage(map.Image,
                    new Rectangle(
                        Convert.ToInt32(map.Area.Left * scaleFactor + horizontalOffset),
                        Convert.ToInt32(map.Area.Top * scaleFactor + verticalOffset),
                        Convert.ToInt32(map.Area.Width * scaleFactor),
                        Convert.ToInt32(map.Area.Height * scaleFactor)),
                    new Rectangle(
                        0,
                        0,
                        Convert.ToInt32(map.Image.Width),
                        Convert.ToInt32(map.Image.Height)),
                        GraphicsUnit.Pixel);
            }
        }

    }
}
